@@
@@ Copyright (C) 2015  Roman Pauer
@@
@@ This library is free software; you can redistribute it and/or
@@ modify it under the terms of the GNU Lesser General Public
@@ License as published by the Free Software Foundation; either
@@ version 2.1 of the License, or (at your option) any later version.
@@
@@ This library is distributed in the hope that it will be useful,
@@ but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
@@ Lesser General Public License for more details.
@@
@@ You should have received a copy of the GNU Lesser General Public
@@ License along with this library; if not, write to the Free Software
@@ Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
@@

#include "hqx-format.h"

#if ((HQX_FORMAT_16 == HQX_FORMAT_RGB565) || (HQX_FORMAT_16 == HQX_FORMAT_BGR565))
    .equ g_sz, (6)
    .equ g_mask, (63 << 2)
#elif ((HQX_FORMAT_16 == HQX_FORMAT_RGB555) || (HQX_FORMAT_16 == HQX_FORMAT_BGR555))
    .equ g_sz, (5)
    .equ g_mask, (31 << 3)
#else
    #error Bad format selected - HQX_FORMAT_16
#endif

#if ((HQX_FORMAT_16 == HQX_FORMAT_RGB565) || (HQX_FORMAT_16 == HQX_FORMAT_RGB555))
    dBB .req d17
    dRR .req d16
    rBB .req r6
    rRR .req r4
#else
    dBB .req d16
    dRR .req d17
    rBB .req r4
    rRR .req r6
#endif

.equ g_sh, (g_sz - 3)
.equ rb_sh_2, (g_sz + 2)

.arm


.global convert_yuv_16
.global convert_yuv_8_rgb_16

.align 4
convert_yuv_16:

@ r0     = const uint16_t *src
@ r1     = uint32_t *dst
@ r2     = unsigned int width
@ lr     = return address

        mov r3, r2                  @ counter = width
        cmp r2, #8                  @ width < 8 ?

        blo 2f

        vmov.i16 d7, #g_mask
        vld1.16 {d16}, [r0]!        @ d16 = src[0-3]; src += 4
        vmov.i16 d6, #(128 << 2)

        sub r3, r3, #4              @ counter -= 4

        vshr.u16 d17, d16, #rb_sh_2 @ d17 = src[0-3] >> 8 -- 'b'
        vshr.u16 d18, d16, #g_sh    @ d18 = src[0-3] >> 3 -- g
        vshl.u16 d16, d16, #3       @ d16 = src[0-3] << 3 -- 'r'

.ifeq (g_sz-6)
        vbic.i16 d17, #0x0007       @ d17 = (src[0-3] >> 8) & (31 << 3) -- 'b'
.else
        vand d17, d17, d7           @ d17 = (src[0-3] >> 8) & (31 << 3) -- 'b'
.endif
        vbic.i16 d16, #0xff00       @ d16 = (src[0-3] << 3) & (31 << 3) -- 'r'
        vand d18, d18, d7           @ d18 = (src[0-3] >> 3) & (63 << 2) -- g

        vmov.i16 d5, #(128 << 3)

    1:
        vshr.u16 q10, q8, #5
        vshr.u16 d19, d18, #g_sz

        vorr q8, q8, q10            @ b |= b >> 5; r |= r >> 5;
        vorr d18, d18, d19          @ g |= g >> 6

        sub r3, r3, #4              @ counter -= 4

        vadd.i16 d29, d6, dRR       @ d29 = (128 << 2) + r -- u
        vadd.i16 d30, d18, d18      @ d30 = 2*g -- v
        vadd.i16 d31, dRR, dBB      @ d31 = r + b
        vadd.i16 d27, d5, d30       @ d27 = (128 << 3) + 2*g -- v
        vsub.i16 d26, d29, dBB      @ d26 = (128 << 2) + r - b -- u
        vsub.i16 d25, d27, d31      @ d25 = (128 << 3) + 2*g - r - b -- v
        vadd.i16 d28, d31, d18      @ d28 = r + g + b -- Y

        vshr.u16 d24, d26, #2       @ d24 = 128 + ((r - b) >> 2) -- u
        vld1.16 {d16}, [r0]!        @ d16 = src[0-3]; src += 4

        vshr.u16 d22, d25, #3       @ d22 = 128 + ((2*g - r - b) >> 3) -- v
        vshr.u16 d23, d28, #2       @ d23 = (r + g + b) >> 2 -- Y

        cmp r3, #4                  @ counter >= 4 ?

        vshr.u16 d17, d16, #rb_sh_2 @ d17 = src[0-3] >> 8 -- 'b'
        vshr.u16 d18, d16, #g_sh    @ d18 = src[0-3] >> 3 -- g
        vshl.u16 d16, d16, #3       @ d16 = src[0-3] << 3 -- 'r'

        vtrn.8 d22, d24             @ d22 = v + (u << 8), d24 = 0

.ifeq (g_sz-6)
        vbic.i16 d17, #0x0007       @ d17 = (src[0-3] >> 8) & (31 << 3) -- 'b'
.else
        vand d17, d17, d7           @ d17 = (src[0-3] >> 8) & (31 << 3) -- 'b'
.endif
        vbic.i16 d16, #0xff00       @ d16 = (src[0-3] << 3) & (31 << 3) -- 'r'
        vand d18, d18, d7           @ d18 = (src[0-3] >> 3) & (63 << 2) -- g

        vst2.16 {d22,d23}, [r1:64]! @ dst[0-3] = v + (u << 8) + (Y << 16), dst+= 4

        bhs 1b

        vshr.u16 q10, q8, #5
        vshr.u16 d19, d18, #g_sz

        vorr q8, q8, q10            @ b |= b >> 5; r |= r >> 5;
        vorr d18, d18, d19          @ g |= g >> 6

        vadd.i16 d29, d6, dRR       @ d29 = (128 << 2) + r -- u
        vadd.i16 d30, d18, d18      @ d30 = 2*g -- v
        vadd.i16 d31, dRR, dBB      @ d31 = r + b
        vadd.i16 d27, d5, d30       @ d27 = (128 << 3) + 2*g -- v
        vsub.i16 d26, d29, dBB      @ d26 = (128 << 2) + r - b -- u
        vsub.i16 d25, d27, d31      @ d25 = (128 << 3) + 2*g - r - b -- v
        vadd.i16 d28, d31, d18      @ d28 = r + g + b -- Y

        vshr.u16 d24, d26, #2       @ d24 = 128 + ((r - b) >> 2) -- u
        vshr.u16 d22, d25, #3       @ d22 = 128 + ((2*g - r - b) >> 3) -- v
        vshr.u16 d23, d28, #2       @ d23 = (r + g + b) >> 2 -- Y

        vtrn.8 d22, d24             @ d22 = v + (u << 8), d24 = 0

        cmp r3, #0                  @ counter != 0 ?

        vst2.16 {d22,d23}, [r1:64]! @ dst[0-3] = v + (u << 8) + (Y << 16), dst+= 4

        bne 2f

        sub r2, r1, r2, lsl #2      @ r2 = dst - width

        ldr r3, [r2]                @ r3 = origdst[0]

        vst2.16 {d22[3],d23[3]}, [r1:32]    @ origdst[width] = origdst[width - 1]

        str r3, [r2, #-4]           @ origdst[-1] = origdst[0]
        bx lr

    2:
        push {r4-r7}

    3:
        ldrh r4, [r0], #2           @ r4 = src[0], src++

        mov r6, r4, lsr #rb_sh_2    @ 'b'
        mov r5, r4, lsr #g_sh       @ g

        mov r4, r4, lsl #3          @ 'r'
        and r6, r6, #(31 << 3)      @ 'b'

        uxtb r4, r4                 @ 'r'
        and r5, r5, #g_mask         @ g

        orr r6, r6, r6, lsr #5      @ 'b'
        orr r4, r4, r4, lsr #5      @ 'r'

        orr r5, r5, r5, lsr #g_sz   @ g
        add r7, rRR, #(128 << 2)    @ r7 = (128 << 2) + r -- u

        add ip, r5, r5              @ ip = 2*g -- v
        add rRR, rRR, rBB           @ rRR = r + b

        add ip, ip, #(128 << 3)     @ ip = (128 << 3) + 2*g -- v
        sub r7, r7, rBB             @ r7 = (128 << 2) + r - b -- u

        sub ip, ip, rRR             @ ip = (128 << 3) + 2*g - r - b -- v
        add r5, rRR, r5             @ r5 = r + g + b -- Y

        mov r7, r7, lsr #2          @ r7 = 128 + ((r - b) >> 2) -- u

        mov ip, ip, lsr #3          @ ip = 128 + ((2*g - r - b) >> 3) -- v
        mov r5, r5, lsr #2          @ r5 = (r + g + b) >> 2 -- Y

        orr ip, ip, r7, lsl #8      @ ip = v + (u << 8)
        subS r3, r3, #1             @ counter -= 1

        orr ip, ip, r5, lsl #16     @ ip = v + (u << 8) + (Y >> 16)
        str ip, [r1], #4            @ dst[0] = ip, dst++

        bne 3b

        sub r2, r1, r2, lsl #2      @ r2 = dst - width

        ldr r3, [r2]                @ r3 = origdst[0]

        str ip, [r1]                @ origdst[width] = origdst[width - 1]

        str r3, [r2, #-4]           @ origdst[-1] = origdst[0]

        pop {r4-r7}
        bx lr

@ end procedure convert_yuv_16


convert_yuv_8_rgb_16:

@ r0     = const uint8_t *src
@ r1     = uint32_t *yuvdst
@ r2     = uint16_t *rgbdst
@ r3     = const uint16_t *pal
@ [sp]   = unsigned int width
@ lr     = return address

        ldr ip, [sp]                @ ip = width
        push {r4-r9}

        mov r8, ip                  @ counter = width
        cmp ip, #8                  @ width < 8 ?

        blo 2f

        ldr r4, [r0], #4            @ r4 = src[0-3], src += 4

        ubfx r7, r4, #0, #8         @ r7 = src[0]
        ubfx r6, r4, #8, #8         @ r6 = src[1]

        ldr r7, [r3, r7, lsl #2]    @ r7 = pal[src[0]]
        ubfx r5, r4, #16, #8        @ r5 = src[2]

        ldr r6, [r3, r6, lsl #2]    @ r6 = pal[src[1]]
        mov r4, r4, lsr #24         @ r4 = src[3]

        ldr r5, [r3, r5, lsl #2]    @ r5 = pal[src[2]]

        ldr r4, [r3, r4, lsl #2]    @ r4 = pal[src[3]]

        bfi r7, r6, #16, #16        @ r7 = pal[src[0]] | pal[src[1]] << 16
        vmov.i16 d7, #g_mask

        bfi r5, r4, #16, #16        @ r5 = pal[src[0]] | pal[src[1]] << 16
        vmov.i16 d6, #(128 << 2)

        vmov d16, r7, r5            @ d16 = pal[src[0-3]]
        sub r8, r8, #4              @ counter -= 4

        vmov.i16 d5, #(128 << 3)

        vshr.u16 d17, d16, #rb_sh_2 @ d17 = src[0-3] >> 8 -- 'b'
        vshr.u16 d18, d16, #g_sh    @ d18 = src[0-3] >> 3 -- g
        vst1.16 {d16}, [r2:64]!     @ rgbdst[0-3] = pal[src[0-3]]; rgbdst += 4
        vshl.u16 d16, d16, #3       @ d16 = src[0-3] << 3 -- 'r'

.ifeq (g_sz-6)
        vbic.i16 d17, #0x0007       @ d17 = (src[0-3] >> 8) & (31 << 3) -- 'b'
.else
        vand d17, d17, d7           @ d17 = (src[0-3] >> 8) & (31 << 3) -- 'b'
.endif
        vbic.i16 d16, #0xff00       @ d16 = (src[0-3] << 3) & (31 << 3) -- 'r'
        vand d18, d18, d7           @ d18 = (src[0-3] >> 3) & (63 << 2) -- g

    1:
        ldr r4, [r0], #4            @ r4 = src[0-3], src += 4

        vshr.u16 q10, q8, #5
        vshr.u16 d19, d18, #g_sz

        vorr q8, q8, q10            @ b |= b >> 5; r |= r >> 5;
        vorr d18, d18, d19          @ g |= g >> 6

        sub r8, r8, #4              @ counter -= 4

        ubfx r7, r4, #0, #8         @ r7 = src[0]
        vadd.i16 d29, d6, dRR       @ d29 = (128 << 2) + r -- u
        ubfx r6, r4, #8, #8         @ r6 = src[1]
        vadd.i16 d30, d18, d18      @ d30 = 2*g -- v

        ldr r7, [r3, r7, lsl #2]    @ r7 = pal[src[0]]
        vadd.i16 d31, dRR, dBB      @ d31 = r + b
        ubfx r5, r4, #16, #8        @ r5 = src[2]
        vadd.i16 d27, d5, d30       @ d27 = (128 << 3) + 2*g -- v

        ldr r6, [r3, r6, lsl #2]    @ r6 = pal[src[1]]
        vsub.i16 d26, d29, dBB      @ d26 = (128 << 2) + r - b -- u
        mov r4, r4, lsr #24         @ r4 = src[3]
        vsub.i16 d25, d27, d31      @ d25 = (128 << 3) + 2*g - r - b -- v

        ldr r5, [r3, r5, lsl #2]    @ r5 = pal[src[2]]
        vadd.i16 d28, d31, d18      @ d28 = r + g + b -- Y

        ldr r4, [r3, r4, lsl #2]    @ r4 = pal[src[3]]
        vshr.u16 d24, d26, #2       @ d24 = 128 + ((r - b) >> 2) -- u

        bfi r7, r6, #16, #16        @ r7 = pal[src[0]] | pal[src[1]] << 16
        vshr.u16 d22, d25, #3       @ d22 = 128 + ((2*g - r - b) >> 3) -- v

        bfi r5, r4, #16, #16        @ r5 = pal[src[0]] | pal[src[1]] << 16
        vshr.u16 d23, d28, #2       @ d23 = (r + g + b) >> 2 -- Y

        vmov d16, r7, r5            @ d16 = pal[src[0-3]]

        cmp r8, #4                  @ counter >= 4 ?

        vshr.u16 d17, d16, #rb_sh_2 @ d17 = src[0-3] >> 8 -- 'b'
        vshr.u16 d18, d16, #g_sh    @ d18 = src[0-3] >> 3 -- g
        vst1.16 {d16}, [r2:64]!     @ rgbdst[0-3] = pal[src[0-3]]; rgbdst += 4
        vshl.u16 d16, d16, #3       @ d16 = src[0-3] << 3 -- 'r'

        vtrn.8 d22, d24             @ d22 = v + (u << 8), d24 = 0

.ifeq (g_sz-6)
        vbic.i16 d17, #0x0007       @ d17 = (src[0-3] >> 8) & (31 << 3) -- 'b'
.else
        vand d17, d17, d7           @ d17 = (src[0-3] >> 8) & (31 << 3) -- 'b'
.endif
        vbic.i16 d16, #0xff00       @ d16 = (src[0-3] << 3) & (31 << 3) -- 'r'
        vand d18, d18, d7           @ d18 = (src[0-3] >> 3) & (63 << 2) -- g

        vst2.16 {d22,d23}, [r1:64]! @ yuvdst[0-3] = v + (u << 8) + (Y << 16), yuvdst+= 4

        bhs 1b

        vshr.u16 q10, q8, #5
        vshr.u16 d19, d18, #g_sz

        vorr q8, q8, q10            @ b |= b >> 5; r |= r >> 5;
        vorr d18, d18, d19          @ g |= g >> 6

        vadd.i16 d29, d6, dRR       @ d29 = (128 << 2) + r -- u
        vadd.i16 d30, d18, d18      @ d30 = 2*g -- v
        vadd.i16 d31, dRR, dBB      @ d31 = r + b
        vadd.i16 d27, d5, d30       @ d27 = (128 << 3) + 2*g -- v
        vsub.i16 d26, d29, dBB      @ d26 = (128 << 2) + r - b -- u
        vsub.i16 d25, d27, d31      @ d25 = (128 << 3) + 2*g - r - b -- v
        vadd.i16 d28, d31, d18      @ d28 = r + g + b -- Y

        vshr.u16 d24, d26, #2       @ d24 = 128 + ((r - b) >> 2) -- u
        vshr.u16 d22, d25, #3       @ d22 = 128 + ((2*g - r - b) >> 3) -- v
        vshr.u16 d23, d28, #2       @ d23 = (r + g + b) >> 2 -- Y

        vtrn.8 d22, d24             @ d22 = v + (u << 8), d24 = 0

        cmp r8, #0                  @ counter != 0 ?

        vst2.16 {d22,d23}, [r1:64]! @ yuvdst[0-3] = v + (u << 8) + (Y << 16), yuvdst+= 4

        bne 2f

        sub ip, r1, ip, lsl #2      @ ip = yuvdst - width

        ldr r3, [ip]                @ r3 = origdst[0]

        vst2.16 {d22[3],d23[3]}, [r1:32]    @ origdst[width] = origdst[width - 1]

        str r3, [ip, #-4]           @ origdst[-1] = origdst[0]

        pop {r4-r9}
        bx lr

    2:
        ldrb r4, [r0], #1           @ r4 = src[0], src++

        ldr r9, [r3, r4, lsl #2]    @ r9 = pal[src[0]]

        mov r6, r9, lsr #rb_sh_2    @ 'b'
        mov r5, r9, lsr #g_sh       @ g

        mov r4, r9, lsl #3          @ 'r'
        and r6, r6, #(31 << 3)      @ 'b'

        uxtb r4, r4                 @ 'r'
        and r5, r5, #g_mask         @ g

        orr r6, r6, r6, lsr #5      @ 'b'
        orr r4, r4, r4, lsr #5      @ 'r'

        orr r5, r5, r5, lsr #g_sz   @ g
        strh r9, [r2], #2           @ rgbdst[0] = r9, rgbdst++

        add r7, rRR, #(128 << 2)    @ r7 = (128 << 2) + r -- u
        add r9, r5, r5              @ r9 = 2*g -- v

        add rRR, rRR, rBB           @ rRR = r + b
        add r9, r9, #(128 << 3)     @ r9 = (128 << 3) + 2*g -- v

        sub r7, r7, rBB             @ r7 = (128 << 2) + r - b -- u
        sub r9, r9, rRR             @ r9 = (128 << 3) + 2*g - r - b -- v

        add r5, rRR, r5             @ r5 = r + g + b -- Y

        mov r7, r7, lsr #2          @ r7 = 128 + ((r - b) >> 2) -- u
        mov r9, r9, lsr #3          @ r9 = 128 + ((2*g - r - b) >> 3) -- v

        mov r5, r5, lsr #2          @ r5 = (r + g + b) >> 2 -- Y
        orr r9, r9, r7, lsl #8      @ r9 = v + (u << 8)

        subS r8, r8, #1             @ counter -= 1
        orr r9, r9, r5, lsl #16     @ ip = v + (u << 8) + (Y >> 16)

        str r9, [r1], #4            @ yuvdst[0] = r9, yuvdst++
        bne 2b

        sub ip, r1, ip, lsl #2      @ ip = yuvdst - width

        ldr r3, [ip]                @ r3 = origyuvdst[0]

        str r9, [r1]                @ origyuvdst[width] = origyuvdst[width - 1]

        str r3, [ip, #-4]           @ origyuvdst[-1] = origyuvdst[0]

        pop {r4-r9}
        bx lr

@ end procedure convert_yuv_8_rgb_16


