/*
 * Copyright (C) 2003  MaxSt ( maxst@hiend3d.com )
 *
 * Copyright (C) 2015  Roman Pauer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __HQX_INTERP_H_
#define __HQX_INTERP_H_

#include <stdint.h>
#include "hqx-format.h"

/* Interpolate functions */
#if ((HQX_FORMAT_16 == HQX_FORMAT_RGB565) || (HQX_FORMAT_16 == HQX_FORMAT_BGR565))
    #define INTERPOLATION_MASK 0x07e0f81f
#elif ((HQX_FORMAT_16 == HQX_FORMAT_RGB555) || (HQX_FORMAT_16 == HQX_FORMAT_BGR555))
    #define INTERPOLATION_MASK 0x03e07c1f
#else
    #error Bad format selected - HQX_FORMAT_16
#endif

#define GROW(n) n = (n | n << 16) & INTERPOLATION_MASK
static uint16_t inline pack(uint32_t n)
{
    n &= INTERPOLATION_MASK;
    return n | (n >> 16);
}

static uint16_t inline Interp1(uint32_t c1, uint32_t c2)
{
//    *pc = (c1*3+c2) >> 2;
    GROW(c1); GROW(c2);
    return pack((c1 * 3 + c2) >> 2);
}

static uint16_t inline Interp2(uint32_t c1, uint32_t c2, uint32_t c3)
{
//    *pc = (c1*2+c2+c3) >> 2;
    GROW(c1); GROW(c2); GROW(c3);
    return pack((c1 * 2 + c2 + c3) >> 2);
}

static uint16_t inline Interp3(uint32_t c1, uint32_t c2)
{
//    *pc = (c1*7+c2)/8;
    GROW(c1); GROW(c2);
    return pack((c1 * 7 + c2) >> 3);
}

static uint16_t inline Interp4(uint32_t c1, uint32_t c2, uint32_t c3)
{
//    *pc = (c1*2+(c2+c3)*7)/16;
    GROW(c1); GROW(c2); GROW(c3);
    return pack((c1 * 2 + (c2 + c3) * 7) >> 4);
}

static uint16_t inline Interp5(uint32_t c1, uint32_t c2)
{
//    *pc = (c1+c2) >> 1;
    GROW(c1); GROW(c2);
    return pack((c1 + c2) >> 1);
}

static uint16_t inline Interp6(uint32_t c1, uint32_t c2, uint32_t c3)
{
//    *pc = (c1*5+c2*2+c3)/8;
    GROW(c1); GROW(c2); GROW(c3);
    return pack((c1 * 5 + c2 * 2 + c3) >> 3);
}

static uint16_t inline Interp7(uint32_t c1, uint32_t c2, uint32_t c3)
{
//    *pc = (c1*6+c2+c3)/8;
    GROW(c1); GROW(c2); GROW(c3);
    return pack((c1 * 6 + c2 + c3) >> 3);
}

static uint16_t inline Interp8(uint32_t c1, uint32_t c2)
{
//    *pc = (c1*5+c2*3)/8;
    GROW(c1); GROW(c2);
    return pack((c1 * 5 + c2 * 3) >> 3);
}

static uint16_t inline Interp9(uint32_t c1, uint32_t c2, uint32_t c3)
{
//    *pc = (c1*2+(c2+c3)*3)/8;
    GROW(c1); GROW(c2); GROW(c3);
    return pack((c1 * 2 + (c2 + c3) * 3) >> 3);
}

static uint16_t inline Interp10(uint32_t c1, uint32_t c2, uint32_t c3)
{
//    *pc = (c1*14+c2+c3)/16;
    GROW(c1); GROW(c2); GROW(c3);
    return pack((c1 * 14 + c2 + c3) >> 4);
}

#endif

